import posixpath
from pprint import pformat

import yaml
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx import addnodes
from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import nested_parse_with_titles

from .md2rst import md2rst
from .nodes import (
    asdf_ref,
    asdf_tree,
    example_description,
    example_item,
    example_section,
    schema_combiner_body,
    schema_combiner_item,
    schema_combiner_list,
    schema_description,
    schema_doc,
    schema_header_title,
    schema_properties,
    schema_property,
    schema_property_details,
    schema_property_name,
    schema_title,
    section_header,
    toc_link,
)

SCHEMA_DEF_SECTION_TITLE = "Schema Definitions"
EXAMPLE_SECTION_TITLE = "Examples"
INTERNAL_DEFINITIONS_SECTION_TITLE = "Internal Definitions"
ORIGINAL_SCHEMA_SECTION_TITLE = "Original Schema"


class schema_def(nodes.comment):
    def __init__(self, *args, **kwargs):
        self.schema_root = kwargs.pop("schema_root", "")
        self.standard_prefix = kwargs.pop("standard_prefix", "")
        super().__init__(*args, **kwargs)


class AsdfAutoschemas(SphinxDirective):

    required_arguments = 0
    optional_arguments = 2
    has_content = True
    option_spec = {
        "schema_root": directives.path,
        "standard_prefix": directives.unchanged,
    }

    def _process_asdf_toctree(self, standard_prefix):

        links = []
        for name in self.content:
            if not name:
                continue
            schema = self.env.path2doc(name.strip() + ".rst")
            link = posixpath.join("generated", standard_prefix, schema)
            links.append((schema, link))

        tocnode = addnodes.toctree()
        tocnode["includefiles"] = [x[1] for x in links]
        tocnode["entries"] = links
        tocnode["maxdepth"] = -1
        tocnode["glob"] = None

        return [tocnode]

    def run(self):

        standard_prefix = self.options.get("standard_prefix", self.env.config.asdf_schema_standard_prefix)

        # This is the case when we are actually using Sphinx to generate
        # documentation
        if not getattr(self.env, "autoasdf_generate", False):
            return self._process_asdf_toctree(standard_prefix)

        # This case allows us to use docutils to parse input documents during
        # the 'builder-inited' phase so that we can determine which new
        # document need to be created by 'autogenerate_schema_docs'. This seems
        # much cleaner than writing a custom parser to extract the schema
        # information.
        schema_root = self.options.get("schema_root", self.env.config.asdf_schema_path)
        return [
            schema_def(text=c.strip().split()[0], standard_prefix=standard_prefix, schema_root=schema_root)
            for c in self.content
        ]


class AsdfSchema(SphinxDirective):

    has_content = True
    optional_arguments = 2
    option_spec = {
        "schema_root": directives.path,
        "standard_prefix": directives.unchanged,
    }

    def run(self):

        self.envconfig = self.state.document.settings.env.config
        self.schema_name = self.content[0]
        schema_dir = self.options.get("schema_root", self.envconfig.asdf_schema_path)
        standard_prefix = self.options.get("standard_prefix", self.envconfig.asdf_schema_standard_prefix)
        srcdir = self.state.document.settings.env.srcdir

        schema_file = posixpath.join(srcdir, schema_dir, standard_prefix, self.schema_name) + ".yaml"

        with open(schema_file) as ff:
            raw_content = ff.read()
            schema = yaml.safe_load(raw_content)

        title = self._parse_title(schema.get("title", ""), schema_file)

        docnodes = schema_doc()
        docnodes.append(title)

        description = schema.get("description", "")
        if description:
            docnodes.append(schema_header_title(text="Description"))
            docnodes.append(self._parse_description(description, schema_file))

        docnodes.append(schema_header_title(text="Outline"))
        docnodes.append(self._create_toc(schema))

        docnodes.append(section_header(text=SCHEMA_DEF_SECTION_TITLE))
        docnodes.append(self._process_properties(schema, top=True))

        examples = schema.get("examples", [])
        if examples:
            docnodes.append(section_header(text=EXAMPLE_SECTION_TITLE))
            docnodes.append(self._process_examples(examples, schema_file))

        if "definitions" in schema:
            docnodes.append(section_header(text=INTERNAL_DEFINITIONS_SECTION_TITLE))
            for name in schema["definitions"]:
                path = f"definitions-{name}"
                tree = schema["definitions"][name]
                required = schema.get("required", [])
                docnodes.append(self._create_property_node(name, tree, name in required, path=path))

        docnodes.append(section_header(text=ORIGINAL_SCHEMA_SECTION_TITLE))
        docnodes.append(nodes.literal_block(text=raw_content, language="yaml"))

        return [docnodes]

    def _create_toc(self, schema):
        toc = nodes.compound()
        toc.append(toc_link(text=SCHEMA_DEF_SECTION_TITLE))
        if "examples" in schema:
            toc.append(toc_link(text=EXAMPLE_SECTION_TITLE))
        if "definitions" in schema:
            toc.append(toc_link(text=INTERNAL_DEFINITIONS_SECTION_TITLE))
        toc.append(toc_link(text=ORIGINAL_SCHEMA_SECTION_TITLE))
        return toc

    def _markdown_to_nodes(self, text, filename):
        """
        This function is taken from the original schema conversion code written
        by Michael Droetboom.
        """
        rst = ViewList()
        for i, line in enumerate(md2rst(text).split("\n")):
            rst.append(line, filename, i + 1)

        node = nodes.section()
        node.document = self.state.document

        nested_parse_with_titles(self.state, rst, node)

        return node.children

    def _parse_title(self, title, filename):
        nodes = self._markdown_to_nodes(title, filename)
        return schema_title(None, *nodes)

    def _parse_description(self, description, filename):
        nodes = self._markdown_to_nodes(description, filename)
        return schema_description(None, *nodes)

    def _resolve_reference(self, schema_id):
        for mapping in self.envconfig.asdf_schema_reference_mappings:
            if schema_id.startswith(mapping[0]):
                relpath = posixpath.relpath(schema_id, mapping[0]).strip("/")
                if relpath == ".":
                    relpath = ""
                schema_id = posixpath.join(mapping[1], relpath).strip("/")
                break

        if not schema_id.endswith(".html"):
            schema_id += ".html"

        return schema_id

    def _create_reference(self, refname, shorten=False):

        if "#" in refname:
            schema_id, fragment = refname.split("#")
        else:
            schema_id = refname
            fragment = ""

        if schema_id:
            schema_id = self._resolve_reference(schema_id)
        if fragment:
            components = fragment.split("/")
            fragment = "#{}".format("-".join(components[1:]))
            if shorten and not schema_id:
                refname = components[-1]
        elif shorten:
            rename = schema_id

        return refname, schema_id + fragment

    def _create_ref_node(self, ref):
        treenodes = asdf_tree()
        refname, href = self._create_reference(ref)
        treenodes.append(asdf_ref(text=refname, href=href))
        return treenodes

    def _create_enum_node(self, enum_values):
        enum_nodes = nodes.compound()
        enum_nodes.append(nodes.line(text="Only the following values are valid for this node:"))
        markdown = "\n".join([f"* **{val}**" for val in enum_values])
        enum_nodes.extend(self._markdown_to_nodes(markdown, ""))
        return enum_nodes

    def _create_array_items_node(self, items, path):
        path = self._append_to_path(path, "items")
        for combiner in ["anyOf", "allOf", "oneOf"]:
            if combiner in items:
                return self._create_combiner(items, combiner, array=True, path=path)

        node_list = nodes.compound()
        if isinstance(items, list):
            text = "The first {} item{} in the list must be the following types:"
            node_list.append(nodes.line(text=text.format(len(items), "s" if len(items) > 1 else "")))
            item_list = nodes.bullet_list()
            for i, it in enumerate(items):
                item_path = self._append_to_path(path, i)
                item_list.append(self._process_properties(it, top=True, path=item_path))
            node_list.append(item_list)
        else:
            node_list.append(nodes.line(text="Items in the array are restricted to the following types:"))
            node_list.append(self._process_properties(items, top=True, path=path))
        return node_list

    def _process_validation_keywords(self, schema, typename=None, path=""):
        node_list = []
        typename = typename or schema["type"]

        if typename == "string":
            if not ("minLength" in schema or "maxLength" in schema):
                node_list.append(nodes.emphasis(text="No length restriction"))
            if schema.get("minLength", 0):
                text = "Minimum length: {}".format(schema["minLength"])
                node_list.append(nodes.line(text=text))
            if "maxLength" in schema:
                text = "Maximum length: {}".format(schema["maxLength"])
                node_list.append(nodes.line(text=text))
            if "pattern" in schema:
                node_list.append(nodes.line(text="Must match the following pattern:"))
                node_list.append(nodes.literal_block(text=schema["pattern"], language="none"))

        elif typename == "array":
            if schema.get("minItems", 0):
                text = "Minimum length: {}".format(schema["minItems"])
                node_list.append(nodes.line(text=text))
            if "maxItems" in schema:
                text = "Maximum length: {}".format(schema["maxItems"])
                node_list.append(nodes.line(text=text))
            if "additionalItems" in schema and "items" in schema:
                if isinstance(schema["items"], list) and schema["additionalItems"] == False:
                    node_list.append(nodes.emphasis(text="Additional items not permitted"))
            elif not ("minItems" in schema or "maxItems" in schema):
                node_list.append(nodes.emphasis(text="No length restriction"))

            if "items" in schema:
                node_list.append(self._create_array_items_node(schema["items"], path=path))

        # TODO: more numerical validation keywords
        elif typename in ["integer", "number"]:
            if "minimum" in schema:
                text = "Minimum value: {}".format(schema["minimum"])
                node_list.append(nodes.line(text=text))
            if "maximum" in schema:
                text = "Maximum value: {}".format(schema["maximum"])
                node_list.append(nodes.line(text=text))

        if "enum" in schema:
            node_list.append(self._create_enum_node(schema["enum"]))

        if "default" in schema:
            if typename in ["string", "integer", "number"]:
                if typename == "string" and not schema["default"]:
                    default = "''"
                else:
                    default = schema["default"]
                text = f"Default value: {default}"
                node_list.append(nodes.line(text=text))
            else:
                default_node = nodes.compound()
                default_node.append(nodes.line(text="Default value:"))
                default_node.append(nodes.literal_block(text=pformat(schema["default"]), language="none"))
                node_list.append(default_node)

        return node_list

    def _process_top_type(self, schema, path=""):
        tree = nodes.compound()
        prop = nodes.compound()
        typename = schema["type"]
        prop.append(schema_property_name(text=typename))
        prop.extend(self._process_validation_keywords(schema, path=path))
        tree.append(prop)
        return tree

    def _append_to_path(self, path, new):
        if not path:
            return str(new).lower()
        else:
            return f"{path}-{new}".lower()

    def _process_properties(self, schema, top=False, path=""):
        for combiner in ["anyOf", "allOf", "oneOf"]:
            if combiner in schema:
                return self._create_combiner(schema, combiner, top=top, path=path)

        if "properties" in schema:
            treenodes = asdf_tree()
            required = schema.get("required", [])
            for key, node in schema["properties"].items():
                new_path = self._append_to_path(path, key)
                treenodes.append(self._create_property_node(key, node, key in required, path=new_path))
            comment = nodes.line(text="This type is an object with the following properties:")
            return schema_properties(None, *[comment, treenodes], id=path)
        elif "type" in schema:
            details = self._process_top_type(schema, path=path)
            return schema_properties(None, details, id=path)
        elif "$ref" in schema:
            ref = self._create_ref_node(schema["$ref"])
            return schema_properties(None, *[ref], id=path)
        elif "tag" in schema:
            ref = self._create_ref_node(schema["tag"])
            return schema_properties(None, *[ref], id=path)
        else:
            text = nodes.emphasis(text="This node has no type definition (unrestricted)")
            return schema_properties(None, text, id=path)

    def _create_combiner(self, items, combiner, array=False, top=False, path=""):
        if top or array:
            container_node = nodes.compound()
        else:
            combiner_path = self._append_to_path(path, "combiner")
            container_node = schema_combiner_body(path=combiner_path)

        path = self._append_to_path(path, combiner)

        if array:
            text = "Items in the array must be **{}** of the following types:"
        else:
            text = "This node must validate against **{}** of the following:"
        text = text.format(combiner.replace("Of", ""))
        text_nodes = self._markdown_to_nodes(text, "")
        container_node.extend(text_nodes)

        combiner_list = schema_combiner_list()
        for i, tree in enumerate(items[combiner]):
            new_path = self._append_to_path(path, i)
            properties = self._process_properties(tree, path=new_path)
            combiner_list.append(schema_combiner_item(None, *[properties]))

        container_node.append(combiner_list)
        container_node["ids"] = [path]
        return schema_properties(None, *[container_node], id=path)

    def _create_property_node(self, name, tree, required, path=""):

        description = tree.get("description", "")

        if "$ref" in tree:
            typ, ref = self._create_reference(tree.get("$ref"), shorten=True)
        elif "tag" in tree:
            _tag = tree.get("tag")
            typ, ref = self._create_reference(_tag, shorten=True)
            if "*" in _tag:
                ref = None
        else:
            typ = tree.get("type", "object")
            ref = None

        prop = schema_property(id=path)
        prop.append(schema_property_name(text=name))
        prop.append(schema_property_details(typ=typ, required=required, ref=ref))
        prop.append(self._parse_description(description, ""))
        if typ != "object":
            prop.extend(self._process_validation_keywords(tree, typename=typ, path=path))
        else:
            prop.append(self._process_properties(tree, path=path))

        prop["ids"] = [path]
        return prop

    def _process_examples(self, tree, filename):
        examples = example_section(num=len(tree))
        for i, example in enumerate(tree):
            node = example_item()
            desc_text = self._markdown_to_nodes(example[0] + ":", filename)
            description = example_description(None, *desc_text)
            node.append(description)
            node.append(nodes.literal_block(text=example[1], language="yaml"))
            examples.append(node)
        return examples



import codecs
import io
import os
import tempfile
import textwrap

import asdf
from asdf import AsdfFile, versioning
from asdf.constants import ASDF_MAGIC, ASDF_STANDARD_COMMENT, BLOCK_FLAG_STREAMED
from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx.util.nodes import set_source_info

version_string = str(versioning.default_version)
TMPDIR = tempfile.mkdtemp()
GLOBALS = {}
FLAGS = {BLOCK_FLAG_STREAMED: "BLOCK_FLAG_STREAMED"}


class RunCodeDirective(Directive):
    has_content = True
    optional_arguments = 1

    def run(self):
        code = textwrap.dedent("\n".join(self.content))

        cwd = os.getcwd()
        os.chdir(TMPDIR)

        try:
            try:
                exec(code, GLOBALS)
            except Exception:
                print(code)
                raise

            literal = nodes.literal_block(code, code)
            literal["language"] = "python"
            set_source_info(self, literal)
        finally:
            os.chdir(cwd)

        if "hidden" not in self.arguments:
            return [literal]
        else:
            return []


class AsdfDirective(Directive):
    required_arguments = 1
    optional_arguments = 1

    def run(self):
        filename = self.arguments[0]

        cwd = os.getcwd()
        os.chdir(TMPDIR)

        show_header = not ("no_header" in self.arguments)
        show_bocks = not ("no_blocks" in self.arguments)

        parts = []
        try:
            ff = AsdfFile()
            with asdf.open(filename, ignore_unrecognized_tag=True, ignore_missing_extensions=True) as af:
                if af.version is None:
                    asdf_standard_version = version_string
                else:
                    asdf_standard_version = af.version

                file_version = af._file_format_version

            header_comments = (
                f"\n{ASDF_MAGIC.strip().decode('utf-8')} {file_version}\n"
                f"#{ASDF_STANDARD_COMMENT.strip().decode('utf-8')} {asdf_standard_version}\n"
            )

            code = AsdfFile._open_impl(ff, filename, _get_yaml_content=True)
            code = header_comments + code.strip().decode("utf-8")
            code += "\n"
            literal = nodes.literal_block(code, code)
            literal["language"] = "yaml"
            set_source_info(self, literal)
            if show_header:
                parts.append(literal)

            kwargs = dict()
            # Use the ignore_unrecognized_tag parameter as a proxy for both options
            kwargs["ignore_unrecognized_tag"] = "ignore_unrecognized_tag" in self.arguments
            kwargs["ignore_missing_extensions"] = "ignore_unrecognized_tag" in self.arguments

            if show_bocks:
                with asdf.open(filename, **kwargs) as ff:
                    for i, block in enumerate(ff.blocks.internal_blocks):
                        data = codecs.encode(block.data.tobytes(), "hex")
                        if len(data) > 40:
                            data = data[:40] + b"..."
                        allocated = block._allocated
                        size = block._size
                        data_size = block._data_size
                        flags = block._flags

                        if flags & BLOCK_FLAG_STREAMED:
                            allocated = size = data_size = 0

                        lines = []
                        lines.append(f"BLOCK {i}:")

                        human_flags = []
                        for key, val in FLAGS.items():
                            if flags & key:
                                human_flags.append(val)
                        if len(human_flags):
                            lines.append("    flags: {}".format(" | ".join(human_flags)))
                        if block.input_compression:
                            lines.append(f"    compression: {block.input_compression}")
                        lines.append(f"    allocated_size: {allocated}")
                        lines.append(f"    used_size: {size}")
                        lines.append(f"    data_size: {data_size}")
                        lines.append(f"    data: {data}")

                        code = "\n".join(lines)
                        code += "\n"

                        literal = nodes.literal_block(code, code)
                        literal["language"] = "yaml"
                        set_source_info(self, literal)
                        parts.append(literal)

                    internal_blocks = list(ff.blocks.internal_blocks)
                    if len(internal_blocks) and internal_blocks[-1].array_storage != "streamed":
                        buff = io.BytesIO()
                        ff.blocks.write_block_index(buff, ff)
                        block_index = buff.getvalue().decode("utf-8")
                        literal = nodes.literal_block(block_index, block_index)
                        literal["language"] = "yaml"
                        set_source_info(self, literal)
                        parts.append(literal)

        finally:
            os.chdir(cwd)

        result = nodes.admonition()
        textnodes, messages = self.state.inline_text(filename, self.lineno)
        title = nodes.title(filename, "", *textnodes)
        result += title
        result += parts
        return [result]



import logging
import os
import posixpath
import warnings

from docutils import nodes
from sphinx.io import read_doc
from sphinx.util import rst
from sphinx.util.docutils import sphinx_domains
from sphinx.util.fileutil import copy_asset

from .directives import schema_def
from .nodes import schema_doc

TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), "templates")


def find_autoasdf_directives(env, filename):
    if filename.endswith(".md"):
        return []

    docname = env.path2doc(filename)
    env.prepare_settings(docname)
    with sphinx_domains(env), rst.default_role(docname, env.config.default_role):
        doctree = read_doc(env.app, env, env.doc2path(docname))

    return doctree.traverse(schema_def)


def find_autoschema_references(app, genfiles):

    # We set this environment variable to indicate that the AsdfSchemas
    # directive should be parsed as a simple list of schema references
    # rather than as the toctree that will be generated when the documentation
    # is actually built.
    app.env.autoasdf_generate = True

    logger = logging.getLogger("sphinx")
    orig_level = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)

    schemas = set()
    for fn in genfiles:
        # Create documentation files based on contents of asdf-schema directives
        path = posixpath.join(app.env.srcdir, fn)
        app.env.temp_data["docname"] = app.env.path2doc(path)
        schemas = schemas.union(find_autoasdf_directives(app.env, path))

    logger.setLevel(orig_level)

    # Unset this variable now that we're done.
    app.env.autoasdf_generate = False

    return list(schemas)


def create_schema_docs(app, schemas):

    for schema in schemas:
        schema_name = schema.children[0].astext()
        standard_prefix = schema.standard_prefix or app.env.config.asdf_schema_standard_prefix
        output_dir = posixpath.join(app.srcdir, "generated", standard_prefix)
        doc_path = posixpath.join(output_dir, schema_name + ".rst")

        if posixpath.exists(doc_path):
            continue

        os.makedirs(posixpath.dirname(doc_path), exist_ok=True)

        with open(doc_path, "w") as ff:
            ff.write(f".. _{standard_prefix}/{schema_name}:\n\n")
            ff.write(schema_name + "\n")
            ff.write("=" * len(schema_name) + "\n\n")
            ff.write(".. asdf-schema::\n")
            if standard_prefix:
                ff.write(f"    :standard_prefix: {standard_prefix}\n")
            ff.write(f"    :schema_root: {schema.schema_root}\n\n")
            ff.write(f"    {schema_name}\n")


def autogenerate_schema_docs(app):

    env = app.env

    genfiles = [env.doc2path(x, base=None) for x in env.found_docs if posixpath.isfile(env.doc2path(x))]

    if not genfiles:
        return

    ext = list(app.config.source_suffix)
    genfiles = [genfile + (not genfile.endswith(tuple(ext)) and ext[0] or "") for genfile in genfiles]

    # Read all source documentation files and parse all asdf-schema directives
    schemas = find_autoschema_references(app, genfiles)
    # Create the documentation files that correspond to the schemas listed
    create_schema_docs(app, schemas)


def update_app_config(app, config):
    from pkg_resources import get_distribution

    dist = get_distribution("sphinx_asdf")
    config.html_context["sphinx_asdf_version"] = dist.version


def handle_page_context(app, pagename, templatename, ctx, doctree):
    # Use custom template when rendering pages containing schema documentation.
    # This allows us to selectively include bootstrap
    if doctree is not None and doctree.traverse(schema_doc):
        return os.path.join(TEMPLATE_PATH, "schema.html")


def normalize_name(name):
    for char in [".", "_", "/"]:
        name = name.replace(char, "-")
    return name


def add_labels_to_nodes(app, document):
    labels = app.env.domaindata["std"]["labels"]
    anonlabels = app.env.domaindata["std"]["anonlabels"]
    basepath = os.path.join("generated", app.env.config.asdf_schema_standard_prefix)

    for node in document.traverse():
        if isinstance(node, str) or not (isinstance(node, nodes.Node) and node["ids"]):
            continue

        labelid = node["ids"][0]
        docname = app.env.docname
        basename = os.path.relpath(docname, basepath)

        if labelid == normalize_name(basename):
            name = basename
        else:
            name = nodes.fully_normalize_name(basename + ":" + labelid)

        # labelname -> docname, labelid
        anonlabels[name] = docname, labelid
        # labelname -> docname, labelid, sectionname
        labels[name] = docname, labelid, ""


def on_build_finished(app, exc):
    if exc is None:
        for asset in ["sphinx_asdf.css", "sphinx_asdf.js"]:
            src = posixpath.join(posixpath.dirname(__file__), "static", asset)
            dst = posixpath.join(app.outdir, "_static")
            copy_asset(src, dst)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Implements a markdown to reST converter using the mistune markdown
parser.

This module was written by Michael Droetboom and has been copied from
http://github.com/spacetelescope/asdf-standard. See
../licenses/ASDF_STANDARD_LICENSE.rst for license information.
"""

import re
import textwrap

import mistune


class BlockLexer(mistune.BlockLexer):
    # Adds math support

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enable_math()

    def enable_math(self):
        self.rules.block_math = re.compile(r"^\$\$(.*?)\$\$", re.DOTALL)
        self.default_rules = ["block_math"] + mistune.BlockLexer.default_rules

    def parse_block_math(self, m):
        self.tokens.append({"type": "block_math", "text": m.group(1)})


class Markdown(mistune.Markdown):
    def output_block_math(self):
        return self.renderer.block_math(self.token["text"])


class InlineLexer(mistune.InlineLexer):
    # Adds math support

    default_rules = ["math"] + mistune.InlineLexer.default_rules

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enable_math()

    def enable_math(self):
        self.rules.text = re.compile(r"^[\s\S]+?(?=[\\<!\[_*`~$]|https?://| {2,}\n|$)")
        self.rules.math = re.compile(r"^\$(.+?)\$")
        self.default_rules = ["math"] + mistune.InlineLexer.default_rules

    def output_math(self, m):
        return self.renderer.math(m.group(1))


class RstRenderer:
    """The default HTML renderer for rendering Markdown."""

    def __init__(self, **kwargs):
        self.options = kwargs
        self.levels = kwargs.get("levels", '*=-^"+:~')

    def _indented(self, content):
        content = textwrap.dedent(content)
        return "\n".join("   " + line for line in content.split("\n"))

    def _directive(self, name, content, args=[], attributes={}):
        out = ".. {}:: {}\n".format(name, ", ".join(args))
        for key, val in attributes.items():
            out += f"    :{key}: {val}\n"
        out += "\n"
        out += self._indented(content)
        out += "\n\n"
        return out

    def placeholder(self):
        return ""

    def block_code(self, code, lang=None):
        code = code.rstrip("\n")
        if lang:
            langs = [lang]
        else:
            langs = ["none"]
        return self._directive("code-block", code, langs)

    def block_quote(self, text):
        return self._indented(text)

    def block_html(self, html):
        if self.options.get("skip_html"):
            return ""
        return self._directive("raw", html, ["html"])

    def header(self, text, level, raw=None):
        return f"{text}\n{self.levels[level] * len(text)}\n\n"

    def hrule(self):
        return "\n\n--------\n\n"

    def list(self, body, ordered=True):
        if ordered:
            body = ("\n" + body).replace("\n- ", "\n#.")
        return body

    def list_item(self, text):
        return textwrap.fill(text, initial_indent="-  ", subsequent_indent="   ") + "\n\n"

    def paragraph(self, text):
        return "%s\n\n" % text

    def table(self, header, body):
        raise NotImplementedError()

    def table_row(self, content):
        raise NotImplementedError()

    def table_cell(self, content, **flags):
        raise NotImplementedError()

    def double_emphasis(self, text):
        return "**%s**" % text

    def emphasis(self, text):
        return "*%s*" % text

    def codespan(self, text):
        return ":code:`%s`" % text

    def linebreak(self):
        return ""

    def strikethrough(self, text):
        return text

    def text(self, text):
        return text

    def autolink(self, link, is_email=False):
        text = link
        if is_email:
            link = "mailto:%s" % link
        if link.startswith("ref:"):
            return f":ref:`{text} <{link[4:]}>`"
        return f"`{text} <{link}>`__"

    def link(self, link, title, text):
        if link.startswith("javascript:"):
            link = ""
        if link.startswith("ref:"):
            return f":ref:`{text} <{link[4:]}>`"
        return f"`{text} <{link}>`__"

    def image(self, src, title, text):
        if src.startswith("javascript:"):
            src = ""
        options = {}
        if text:
            options["alt"] = text
        return self._directive("image", "", [src], options)

    def tag(self, html):
        if self.options.get("skip_html"):
            return ""
        return ":raw:`%s`" % html

    def newline(self):
        return ""

    def footnote_ref(self, key, index):
        raise NotImplementedError()

    def footnote_item(self, key, text):
        raise NotImplementedError()

    def footnotes(self, text):
        raise NotImplementedError()

    def math(self, text):
        return ":math:`%s`" % text

    def block_math(self, text):
        return self._directive("math", text)


def md2rst(content):
    """
    Convert the given string in markdown to a string of reST.
    """
    renderer = RstRenderer()
    md = Markdown(block=BlockLexer, inline=InlineLexer, renderer=renderer)
    return md.render(content)



import os
import pickle
from tempfile import gettempdir

import pytest
from docutils import nodes
from sphinx.testing.path import path

from sphinx_asdf import nodes as sa_nodes


@pytest.fixture(scope="session")
def rootdir():
    return path(__file__).parent.abspath() / "roots"


@pytest.fixture(scope="session")
def sphinx_test_tempdir():
    """
    temporary directory that wrapped with `path` class.
    """

    def make_tmpdir():
        return os.path.join(gettempdir(), str(os.getpid()))

    return path(os.environ.get("SPHINX_TEST_TEMPDIR", make_tmpdir())).abspath()


@pytest.mark.sphinx("dummy", testroot="basic-generation")
def test_basic_generation(app, status, warning):
    app.builder.build_all()

    foo_doc = app.srcdir / "generated" / "foo.rst"
    assert foo_doc.exists()
    assert (".. asdf-schema::\n" "    :schema_root: schemas\n" "\n" "    foo\n") in foo_doc.read_text()

    bar_doc = app.srcdir / "generated" / "bar.rst"
    assert bar_doc.exists()
    assert (".. asdf-schema::\n" "    :schema_root: schemas\n" "\n" "    bar\n") in bar_doc.read_text()

    baz_doc = app.srcdir / "generated" / "core" / "baz.rst"
    assert baz_doc.exists()
    assert (".. asdf-schema::\n" "    :schema_root: schemas\n" "\n" "    core/baz\n") in baz_doc.read_text()


@pytest.mark.sphinx("dummy", testroot="global-config")
def test_generation_global_config(app, status, warning):
    app.builder.build_all()

    foo_doc = app.srcdir / "generated" / "foo.rst"
    assert foo_doc.exists()
    assert (".. asdf-schema::\n" "    :schema_root: a/b/c/schemas\n" "\n" "    foo\n") in foo_doc.read_text()

    bar_doc = app.srcdir / "generated" / "bar.rst"
    assert bar_doc.exists()
    assert (".. asdf-schema::\n" "    :schema_root: a/b/c/schemas\n" "\n" "    bar\n") in bar_doc.read_text()

    baz_doc = app.srcdir / "generated" / "core" / "baz.rst"
    assert baz_doc.exists()
    assert (".. asdf-schema::\n" "    :schema_root: a/b/c/schemas\n" "\n" "    core/baz\n") in baz_doc.read_text()


@pytest.mark.sphinx("dummy", testroot="global-prefix")
def test_generation_global_prefix(app, status, warning):
    app.builder.build_all()

    foo_doc = app.srcdir / "generated" / "a" / "b" / "c" / "foo.rst"
    assert foo_doc.exists()
    assert (
        ".. asdf-schema::\n" "    :standard_prefix: a/b/c\n" "    :schema_root: schemas\n" "\n" "    foo\n"
    ) in foo_doc.read_text()

    bar_doc = app.srcdir / "generated" / "a" / "b" / "c" / "bar.rst"
    assert bar_doc.exists()
    assert (
        ".. asdf-schema::\n" "    :standard_prefix: a/b/c\n" "    :schema_root: schemas\n" "\n" "    bar\n"
    ) in bar_doc.read_text()

    baz_doc = app.srcdir / "generated" / "a" / "b" / "c" / "core" / "baz.rst"
    assert baz_doc.exists()
    assert (
        ".. asdf-schema::\n" "    :standard_prefix: a/b/c\n" "    :schema_root: schemas\n" "\n" "    core/baz\n"
    ) in baz_doc.read_text()


@pytest.mark.sphinx("dummy", testroot="basic-generation")
def test_basic_build(app, status, warning):
    app.builder.build_all()

    # Test each of the generated schema documents
    for name in ["foo", "bar", "core/baz"]:
        doctree_path = app.doctreedir / "generated" / f"{name}.doctree"
        doc = pickle.loads(doctree_path.read_bytes())

        title = list(doc.traverse(nodes.title))[0]
        assert title.astext() == name

        schema_top = list(doc.traverse(sa_nodes.schema_doc))
        assert len(schema_top) > 0



import os

from sphinx.builders.html import StandaloneHTMLBuilder

from .asdf2rst import AsdfDirective, RunCodeDirective
from .connections import (
    add_labels_to_nodes,
    autogenerate_schema_docs,
    handle_page_context,
    on_build_finished,
    update_app_config,
)
from .directives import AsdfAutoschemas, AsdfSchema, schema_def
from .nodes import add_asdf_nodes


def setup(app):

    # Describes a path relative to the sphinx source directory
    app.add_config_value("asdf_schema_path", "schemas", "env")
    app.add_config_value("asdf_schema_standard_prefix", "", "env")
    app.add_config_value("asdf_schema_reference_mappings", [], "env")

    app.add_directive("asdf-autoschemas", AsdfAutoschemas)
    app.add_directive("asdf-schema", AsdfSchema)
    app.add_directive("runcode", RunCodeDirective)
    app.add_directive("asdf", AsdfDirective)

    add_asdf_nodes(app)

    app.connect("builder-inited", autogenerate_schema_docs)
    app.connect("config-inited", update_app_config)
    app.connect("html-page-context", handle_page_context)
    app.connect("doctree-read", add_labels_to_nodes)
    app.connect("build-finished", on_build_finished)

    static_dir = os.path.join(os.path.dirname(__file__), "static")

    app.config._raw_config.setdefault("html_static_path", []).append(static_dir)
    app.add_css_file("custom.css")

    return dict(version="0.1.1", parallel_read_safe=True, parallel_write_safe=True)



import datetime
import os
import sys
from pathlib import Path

import numpy
import toml

# Ensure documentation examples are determinstically random.
try:
    numpy.random.seed(int(os.environ["SOURCE_DATE_EPOCH"]))
except KeyError:
    pass

try:
    from sphinx_astropy.conf.v1 import *
except ImportError:
    print("ERROR: the documentation requires the sphinx-astropy package to be installed")
    sys.exit(1)

# -- General configuration ----------------------------------------------------
intersphinx_mapping["pypa-packaging"] = ("https://packaging.python.org/en/latest/", None)
intersphinx_mapping["asdf-standard"] = ("https://asdf-standard.readthedocs.io/en/latest/", None)
intersphinx_mapping["pytest"] = ("https://docs.pytest.org/en/latest/", None)

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""

# -- Options for HTML output ---------------------------------------------------
html_theme = "sphinx_rtd_theme"
html_theme_options = {}

static_dir = os.path.join(os.path.dirname(__file__), "static")
html_logo = f"{static_dir}/logo.png"
html_favicon = f"{static_dir}/logo.ico"
latex_logo = f"{static_dir}/logo.pdf"

sys.path.insert(0, os.path.join(os.path.abspath(os.path.dirname("__file__")), "sphinxext"))
extensions += ["sphinx_asdf", "sphinx_inline_tabs", "myst_parser"]
